# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Autogenerated by numba_cuda_mlir/runtime/stubgen.py
# Please do not edit

from numba_cuda_mlir.typing import (
    none,
    ptr,
    int8,
    int16,
    int32,
    int64,
    bool,
    float32,
    float64,
    UniTuple,
    Tuple,
)

def elect_sync() -> bool: ...
def elect_sync_mask(arg0: int32) -> bool: ...
def mbarrier_init(arg0: ptr, arg1: int32) -> None: ...
def mbarrier_try_wait(arg0: ptr, arg1: int32) -> None: ...
def barrier() -> None: ...
def mbarrier_arrive_expect_tx(arg0: ptr, arg1: int32) -> None: ...
def cp_async_bulk_load(arg0: ptr, arg1: ptr, arg2: ptr, arg3: int32) -> None: ...
def fence_mbarrier_init() -> None: ...
def cp_async_bulk_store(arg0: ptr, arg1: ptr, arg2: int32) -> None: ...
def cp_async_bulk_tensor_2d(arg0: ptr, arg1: ptr, arg2: int32, arg3: int32, arg4: ptr) -> None: ...
def cp_async_bulk_tensor_2d_multicast(
    arg0: ptr, arg1: ptr, arg2: int32, arg3: int32, arg4: ptr, arg5: int16
) -> None: ...
def cluster_arrive_relaxed() -> None: ...
def cluster_wait() -> None: ...
def block_idx_in_cluster() -> int32: ...
def cluster_ctarank() -> int32: ...
def mapa(arg0: ptr, arg1: int32) -> ptr: ...
def warp_idx() -> int32: ...
def prefetch_tensormap(arg0: ptr) -> None: ...
def encode_instr_descriptor(
    arg0: int16,
    arg1: bool,
    arg2: bool,
    arg3: int16,
    arg4: bool,
    arg5: int16,
    arg6: int16,
    arg7: bool,
    arg8: bool,
    arg9: bool,
    arg10: bool,
    arg11: int16,
    arg12: int16,
    arg13: int16,
) -> int32: ...
def encode_smem_descriptor(
    arg0: int16, arg1: int16, arg2: int16, arg3: int8, arg4: int8
) -> int64: ...
def breakpoint() -> None: ...
def nanosleep(arg0: int32) -> None: ...
def round_ndigits_type_f32_type_i32(arg0: float32, arg1: int32) -> float32: ...
def round_ndigits_type_f32_type_i64(arg0: float32, arg1: int64) -> float32: ...
def round_ndigits_type_f64_type_i32(arg0: float64, arg1: int32) -> float64: ...
def round_ndigits_type_f64_type_i64(arg0: float64, arg1: int64) -> float64: ...
def __nv_ffs(arg0: int32) -> int32: ...
def __nv_ffsll(arg0: int64) -> int32: ...
def ffs(arg0: int32) -> int32: ...
def ffsll(arg0: int64) -> int32: ...
def tcgen05_alloc(arg0: ptr, arg1: int32) -> None: ...
def tcgen05_mma(arg0: int32, arg1: int64, arg2: int64, arg3: int32, arg4: int32) -> None: ...
def tcgen05_commit_arrive(arg0: ptr) -> None: ...
def tcgen05_ld(arg0: int32) -> float32: ...
def tcgen05_ld_x2(arg0: int32) -> tuple[float32, float32]: ...
def tcgen05_dealloc(arg0: int32, arg1: int32) -> None: ...
def tcgen05_relinquish_alloc_permit() -> None: ...
def tcgen05_alloc_2cta(arg0: ptr, arg1: int32) -> None: ...
def tcgen05_mma_2cta(arg0: int32, arg1: int64, arg2: int64, arg3: int32, arg4: int32) -> None: ...
def tcgen05_dealloc_2cta(arg0: int32, arg1: int32) -> None: ...
def tcgen05_relinquish_alloc_permit_2cta() -> None: ...
