# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.types import float16, boolean


def habs(arg0: float16) -> float16: ...


def hadd(arg0: float16, arg1: float16) -> float16: ...


def hceil(arg0: float16) -> float16: ...


def hcos(arg0: float16) -> float16: ...


def hdiv(arg0: float16, arg1: float16) -> float16: ...


def heq(arg0: float16, arg1: float16) -> boolean: ...


def hexp(arg0: float16) -> float16: ...


def hexp10(arg0: float16) -> float16: ...


def hexp2(arg0: float16) -> float16: ...


def hfloor(arg0: float16) -> float16: ...


def hfma(arg0: float16, arg1: float16, arg2: float16) -> float16: ...


def hge(arg0: float16, arg1: float16) -> boolean: ...


def hgt(arg0: float16, arg1: float16) -> boolean: ...


def hle(arg0: float16, arg1: float16) -> boolean: ...


def hlog(arg0: float16) -> float16: ...


def hlog10(arg0: float16) -> float16: ...


def hlog2(arg0: float16) -> float16: ...


def hlt(arg0: float16, arg1: float16) -> boolean: ...


def hmax(arg0: float16, arg1: float16) -> float16: ...


def hmin(arg0: float16, arg1: float16) -> float16: ...


def hmul(arg0: float16, arg1: float16) -> float16: ...


def hne(arg0: float16, arg1: float16) -> boolean: ...


def hneg(arg0: float16) -> float16: ...


def hrcp(arg0: float16) -> float16: ...


def hrint(arg0: float16) -> float16: ...


def hrsqrt(arg0: float16) -> float16: ...


def hsin(arg0: float16) -> float16: ...


def hsqrt(arg0: float16) -> float16: ...


def hsub(arg0: float16, arg1: float16) -> float16: ...


def htanh(arg0: float16) -> float16: ...


def htanh_approx(arg0: float16) -> float16: ...


def htrunc(arg0: float16) -> float16: ...
